@@ -5,8 +5,10 @@ from decimal import Decimal
 from itertools import groupby, chain
 from functools import partial
 
+from trytond.i18n import gettext
 from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, \
     sequence_ordered
+from trytond.model.exceptions import AccessError
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
@@ -15,7 +17,10 @@ from trytond.transaction import Transaction
 from trytond.pool import Pool
 
 from trytond.modules.account.tax import TaxableMixin
+from trytond.modules.account_product.exceptions import AccountError
 from trytond.modules.product import price_digits
+from .exceptions import (
+    SaleValidationError, SaleQuotationError, PartyLocationError)
 
 __all__ = ['Sale', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
     'SaleLine', 'SaleLineTax', 'SaleLineIgnoredMove',
@@ -175,14 +180,16 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             'readonly': Eval('state') != 'draft',
             },
         depends=['state'])
+    invoice_method_string = invoice_method.translated('invoice_method')
     invoice_state = fields.Selection([
             ('none', 'None'),
             ('waiting', 'Waiting'),
             ('paid', 'Paid'),
             ('exception', 'Exception'),
             ], 'Invoice State', readonly=True, required=True)
-    invoices = fields.Function(fields.One2Many('account.invoice', None,
-            'Invoices'), 'get_invoices', searcher='search_invoices')
+    invoices = fields.Function(fields.Many2Many(
+            'account.invoice', None, None, "Invoices"),
+        'get_invoices', searcher='search_invoices')
     invoices_ignored = fields.Many2Many('sale.sale-ignored-account.invoice',
             'sale', 'invoice', 'Ignored Invoices', readonly=True)
     invoices_recreated = fields.Many2Many(
@@ -197,16 +204,18 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             'readonly': Eval('state') != 'draft',
             },
         depends=['state'])
+    shipment_method_string = shipment_method.translated('shipment_method')
     shipment_state = fields.Selection([
             ('none', 'None'),
             ('waiting', 'Waiting'),
             ('sent', 'Sent'),
             ('exception', 'Exception'),
             ], 'Shipment State', readonly=True, required=True)
-    shipments = fields.Function(fields.One2Many('stock.shipment.out', None,
-        'Shipments'), 'get_shipments', searcher='search_shipments')
-    shipment_returns = fields.Function(
-        fields.One2Many('stock.shipment.out.return', None, 'Shipment Returns'),
+    shipments = fields.Function(fields.Many2Many(
+            'stock.shipment.out', None, None, "Shipments"),
+        'get_shipments', searcher='search_shipments')
+    shipment_returns = fields.Function(fields.Many2Many(
+            'stock.shipment.out.return', None, None, "Shipment Returns"),
         'get_shipment_returns', searcher='search_shipment_returns')
     moves = fields.One2Many('stock.move', 'sale', 'Moves', readonly=True)
     origin = fields.Reference('Origin', selection='get_origin', select=True,
@@ -218,22 +227,10 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     @classmethod
     def __setup__(cls):
         super(Sale, cls).__setup__()
-        cls.create_date.select = True
         cls._order = [
             ('sale_date', 'DESC'),
             ('id', 'DESC'),
             ]
-        cls._error_messages.update({
-                'invalid_method': ('Invalid combination of shipment and '
-                    'invoicing methods on sale "%s".'),
-                'addresses_required': (
-                    'Invoice and Shipment addresses must be '
-                    'defined for the quotation of sale "%s".'),
-                'warehouse_required': ('Warehouse must be defined for the '
-                    'quotation of sale "%s".'),
-                'delete_cancel': ('Sale "%s" must be cancelled before '
-                    'deletion.'),
-                })
         cls._transitions |= set((
                 ('draft', 'quotation'),
                 ('quotation', 'confirmed'),
@@ -270,7 +267,10 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'depends': ['state'],
                     },
                 'process': {
-                    'invisible': Eval('state') != 'confirmed',
+                    'invisible': ~Eval('state').in_(
+                        ['confirmed', 'processing']),
+                    'icon': If(Eval('state') == 'confirmed',
+                        'tryton-forward', 'tryton-refresh'),
                     'depends': ['state'],
                     },
                 'handle_invoice_exception': {
@@ -307,6 +307,8 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         # Migration from 4.0: Drop not null on payment_term
         table.not_null_action('payment_term', 'remove')
 
+        # Add index on create_date
+        table.index_action('create_date', action='add')
 
     @classmethod
     def default_payment_term(cls):
@@ -590,10 +592,18 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         '''
         if (self.invoice_method == 'shipment'
                 and self.shipment_method in ('invoice', 'manual')):
-            self.raise_user_error('invalid_method', (self.rec_name,))
+            raise SaleValidationError(
+                gettext('sale.msg_sale_invalid_method',
+                    invoice_method=self.invoice_method_string,
+                    shipment_method=self.shipment_method_string,
+                    sale=self.rec_name))
         if (self.shipment_method == 'invoice'
                 and self.invoice_method in ('shipment', 'manual')):
-            self.raise_user_error('invalid_method', (self.rec_name,))
+            raise SaleValidationError(
+                gettext('sale.msg_sale_invalid_method',
+                    invoice_method=self.invoice_method_string,
+                    shipment_method=self.shipment_method_string,
+                    sale=self.rec_name))
 
     def get_rec_name(self, name):
         items = []
@@ -649,8 +659,15 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         return super(Sale, cls).copy(sales, default=default)
 
     def check_for_quotation(self):
-        if not self.invoice_address or not self.shipment_address:
-            self.raise_user_error('addresses_required', (self.rec_name,))
+        if not self.invoice_address:
+            raise SaleQuotationError(
+                gettext('sale.msg_sale_invoice_address_required_for_quotation',
+                    sale=self.rec_name))
+        if not self.shipment_address:
+            raise SaleQuotationError(
+                gettext('sale'
+                    '.msg_sale_shipment_address_required_for_quotation',
+                    sale=self.rec_name))
         for line in self.lines:
             if (line.quantity or 0) >= 0:
                 location = line.from_location
@@ -659,8 +676,9 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             if ((not location or not line.warehouse)
                     and line.product
                     and line.product.type in ('goods', 'assets')):
-                self.raise_user_error('warehouse_required',
-                    (self.rec_name,))
+                raise SaleQuotationError(
+                    gettext('sale.msg_sale_warehouse_required_for_quotation',
+                        sale=self.rec_name))
 
     @classmethod
     def set_number(cls, sales):
@@ -815,7 +833,9 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         cls.cancel(sales)
         for sale in sales:
             if sale.state != 'cancel':
-                cls.raise_user_error('delete_cancel', (sale.rec_name,))
+                raise AccessError(
+                    gettext('sale.msg_sale_delete_cancel',
+                        sale=sale.rec_name))
         super(Sale, cls).delete(sales)
 
     @classmethod
@@ -877,7 +897,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     def process(cls, sales):
         done = []
         process = []
-        cls.__lock(sales)
+        cls.lock(sales)
         for sale in sales:
             if sale.state not in ('confirmed', 'processing', 'done'):
                 continue
@@ -898,25 +918,6 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         if done:
             cls.do(done)
 
-    @classmethod
-    def __lock(cls, records):
-        from trytond.tools import grouped_slice, reduce_ids
-        from sql import Literal, For
-        transaction = Transaction()
-        database = transaction.database
-        connection = transaction.connection
-        table = cls.__table__()
-
-        if database.has_select_for():
-            for sub_records in grouped_slice(records):
-                where = reduce_ids(table.id, sub_records)
-                query = table.select(
-                    Literal(1), where=where, for_=For('UPDATE', nowait=True))
-                with connection.cursor() as cursor:
-                    cursor.execute(*query)
-        else:
-            database.lock(connection, cls._table)
-
     @classmethod
     @ModelView.button_action('sale.wizard_modify_header')
     def modify_header(cls, sales):
@@ -986,7 +987,12 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
     product = fields.Many2One('product.product', 'Product',
-        ondelete='RESTRICT', domain=[('salable', '=', True)],
+        ondelete='RESTRICT',
+        domain=[
+            If(Eval('sale_state').in_(['draft', 'quotation']),
+                ('salable', '=', True),
+                ()),
+            ],
         states={
             'invisible': Eval('type') != 'line',
             'readonly': Eval('sale_state') != 'draft',
@@ -1062,21 +1068,6 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     sale_state = fields.Function(fields.Selection(STATES, 'Sale State'),
         'on_change_with_sale_state')
 
-    @classmethod
-    def __setup__(cls):
-        super(SaleLine, cls).__setup__()
-        cls._error_messages.update({
-                'customer_location_required': (
-                    'Sale "%(sale)s" is missing the '
-                    'customer location in line "%(line)s".'),
-                'missing_account_revenue': ('Product "%(product)s" of sale '
-                    '%(sale)s misses a revenue account.'),
-                'missing_default_account_revenue': ('Sale "%(sale)s" '
-                    'misses a default "account revenue".'),
-                'delete_cancel_draft': ('The line "%(line)s" must be on '
-                    'canceled or draft sale to be deleted.'),
-                })
-
     @classmethod
     def __register__(cls, module_name):
         super(SaleLine, cls).__register__(module_name)
@@ -1280,10 +1271,10 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     def on_change_with_shipping_date(self, name=None):
         pool = Pool()
         Date = pool.get('ir.date')
-        moves = [m for m in self.moves if m.state != 'cancel']
-        if moves:
+        if self.moves:
             dates = filter(
-                None, (m.effective_date or m.planned_date for m in moves))
+                None, (m.effective_date or m.planned_date for m in self.moves
+                    if m.state != 'cancel'))
             return min(dates, default=None)
         if self.product and self.quantity is not None and self.quantity > 0:
             date = self.sale.sale_date if self.sale else None
@@ -1308,8 +1299,6 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
 
         invoice_line = InvoiceLine()
         invoice_line.type = self.type
-        invoice_line.currency = self.sale.currency
-        invoice_line.company = self.sale.company
         invoice_line.description = self.description
         invoice_line.note = self.note
         invoice_line.origin = self
@@ -1337,17 +1326,17 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
         if self.product:
             invoice_line.account = self.product.account_revenue_used
             if not invoice_line.account:
-                self.raise_user_error('missing_account_revenue', {
-                        'sale': self.sale.rec_name,
-                        'product': self.product.rec_name,
-                        })
+                raise AccountError(
+                    gettext('sale.msg_sale_product_missing_account_expense',
+                        sale=self.sale.rec_name,
+                        product=self.product.rec_name))
         else:
             invoice_line.account = account_config.get_multivalue(
                 'default_category_account_revenue')
             if not invoice_line.account:
-                self.raise_user_error('missing_default_account_revenue', {
-                        'sale': self.sale.rec_name,
-                        })
+                raise AccountError(
+                    gettext('sale.msg_sale_missing_account_expense',
+                        sale=self.sale.rec_name))
         invoice_line.stock_moves = self._get_invoice_line_moves()
         return [invoice_line]
 
@@ -1428,10 +1417,10 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
             return
 
         if not self.sale.party.customer_location:
-            self.raise_user_error('customer_location_required', {
-                    'sale': self.sale.rec_name,
-                    'line': self.rec_name,
-                    })
+            raise PartyLocationError(
+                gettext('sale.msg_sale_customer_location_required',
+                    sale=self.sale.rec_name,
+                    party=self.sale.party.rec_name))
         move = Move()
         move.quantity = quantity
         move.uom = self.unit
@@ -1517,9 +1506,10 @@ class SaleLine(sequence_ordered(), ModelSQL, ModelView):
     def delete(cls, lines):
         for line in lines:
             if line.sale_state not in {'cancel', 'draft'}:
-                cls.raise_user_error('delete_cancel_draft', {
-                        'line': line.rec_name,
-                        })
+                raise AccessError(
+                    gettext('sale.msg_sale_line_delete_cancel_draft',
+                        line=line.rec_name,
+                        sale=line.sale.rec_name))
         super(SaleLine, cls).delete(lines)
 
     @classmethod
@@ -1784,23 +1774,15 @@ class ModifyHeader(Wizard):
             ])
     modify = StateTransition()
 
-    @classmethod
-    def __setup__(cls):
-        super(ModifyHeader, cls).__setup__()
-        cls._error_messages.update({
-                'not_in_draft': (
-                    'The sale "%(sale)s" must be in draft to modify header.'),
-                })
-
     def get_sale(self):
         pool = Pool()
         Sale = pool.get('sale.sale')
 
         sale = Sale(Transaction().context['active_id'])
         if sale.state != 'draft':
-            self.raise_user_error('not_in_draft', {
-                    'sale': sale.rec_name,
-                    })
+            raise AccessError(
+                gettext('sale.msg_sale_modify_header_draft',
+                    sale=sale.rec_name))
         return sale
 
     def default_start(self, fields):
